%{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
CT 2/2012 - this function turns a 1 layer mfa into a 2 layer mfa 
but will collapse it
PURPOSE:
INPUT:  params_dmfa.prune - wheter or not to prune the model. -1 for no pruning
        params_dmfa.nSamples - how many posterior samples do we take per FA
        params_dmfa.smart_C2 - instead of params_dmfa.C2 per 1st layer
        component, we assign in a smart way
        
OUTPUT: mfacollapsed has non EYE covariance on z
        mfacollapsed2 has EYE covariance on z
        mfamodel2 - pruned model
        mfalayer2 - cell, 1 mfa for each 1st layer FA
NOTES:          
TESTED:
CHANGELOG:
TODO:

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}
function [mfacollapsed mfacollapsed2 mfamodel2 mfalayer2] = ...
          deepen_mfa( mfamodel, params_dmfa)

global xxTrain;
global xxTest;
global xxValid;
xxValid = [];

assert(length(size(xxTrain)) == 2 && length(size(xxTest)) == 2);

Dim = size(xxTrain,2);
assert(Dim == size(xxTest,2));

%%
C = size(mfamodel.M,2);
d = size(mfamodel.W,2);
N = size(xxTrain,1);
goodcomponents = find(mfamodel.mix > params_dmfa.prune);

mfamodel2 = mfamodel;
mfamodel2.mix = mfamodel2.mix(goodcomponents);
mfamodel2.W = mfamodel2.W(:,:,goodcomponents);
mfamodel2.M = mfamodel2.M(:,goodcomponents);
mfamodel2.Psi = mfamodel2.Psi(:,goodcomponents);
mfamodel2.mix = mfamodel2.mix/sum(mfamodel2.mix);

likv = eval_mfa( double(xxTest'), mfamodel);
orig_testlogp = mean(likv);

%test smaller model
likv = eval_mfa( double(xxTest'), mfamodel2);
prune_testlogp = mean(likv);

mfamodel
mfamodel2

fprintf('\nDeepMFA: orig test:%f prune test:%f\n',orig_testlogp, prune_testlogp);


%% get data for 2nd layer
clear xxTrain2Data;
clear xxTrain2;

C = size(mfamodel2.M,2);
d = size(mfamodel2.W,2);

nSamples = params_dmfa.nSamples;  %enlarge the dataset by nSamples times

if size(xxTrain,1) > 10000
    xxTrain = batchdata_reshape(xxTrain, [size(xxTrain,1)/10 size(xxTrain,2) 10]);
end

for bb = 1:size(xxTrain,3)    
    fprintf(' %.2f', bb/double(size(xxTrain,3)));
    
    [LogPxGc, Lats, lat_covs] = ll_mfa( double(xxTrain(:,:,bb)'), mfamodel2, double(xxTrain(:,:,bb)').^2, 1);
    assert(all(isfinite(LogPxGc(:))));
    LogPxc = bsxfun(@plus, LogPxGc, log(mfamodel2.mix+realmin));
    Q    = bsxfun(@minus, LogPxc, max(LogPxc,[],1) );
    Q    = exp(Q)+realmin;
    Q    = bsxfun(@rdivide, Q , sum(Q,1) );
    
    [val Q_which_c] = max(Q, [], 1); %make a hard choice!   
   
    for c = 1:C
        inds = find(Q_which_c == c);
        assert(length(inds) > 0);
        Lats2 = Lats(:, inds, c);
        
        L = chol(lat_covs(:,:,c), 'lower');
        xxTrain2{c,bb} = repmat(Lats2, 1, nSamples);
        xxTrain2{c,bb} =  xxTrain2{c,bb} + L*randn(d, nSamples*size(Lats2,2) ); %add some posterior noise
    end
    
end

xxTrain = batchdata_reshape(xxTrain);

for c = 1:C
    xxTrain2Data{c} = cat(2,xxTrain2{c,:});
end
clear Lats;

% ----- test data -----
[LogPxGc, Lats, lat_covs] = ll_mfa( double(xxTest'), mfamodel2, double( xxTest' ).^2, 1);
assert(all(isfinite(LogPxGc(:))));
LogPxc = bsxfun(@plus, LogPxGc, log(mfamodel2.mix+realmin));
Q    = bsxfun(@minus, LogPxc, max(LogPxc,[],1) );
Q    = exp(Q)+realmin;
Q    = bsxfun(@rdivide, Q , sum(Q,1) );

[val Q_which_c] = max(Q, [], 1); %make a hard choice!

xxTest2Data =[];

nSamples = 1;  %enlarge the dataset by nSamples times

for c = 1:C
    inds = find(Q_which_c == c);
    if (length(inds) <= 0)
       warning('xxTest has no data belong to class c');
    else
        Lats2 = Lats(:, inds, c);
        L = chol(lat_covs(:,:,c), 'lower');
        xxTest2Data{c} = repmat(Lats2, 1, nSamples);
        xxTest2Data{c} = xxTest2Data{c} + L*randn(d, nSamples*size(Lats2,2) );         
    end
end

clear Lats;

if params_dmfa.smart_C2 ==0    
    C2 = ones(1, C)*params_dmfa.C2;
else    
    %[gll_val gll_ind] = sort( train_gauss_ll);
    
    [gll_val gll_ind] = sort( mfamodel2.mix,1,'descend');
    total_C2 = params_dmfa.C2*C;
    
    nums = int32(linspace(total_C2/C*2-2, 2, C));
    C2 = [];
    C2(gll_ind) = nums;
    [C2; mfamodel2.mix'];
end


%% learn 2nd layer
d2 = params_dmfa.d2;
assert( d2 <= d);
mfalayer2 = cell(1,C);  % cell to store a MFA for each 1st layer FA

for c = 1:C
    SamplesPerBatch = 100000;
    totalsize = double(size(xxTrain2Data{c},2));
    nBatches = double(max(1, int32(floor( totalsize/SamplesPerBatch)) ));
    nSamples = int32(floor( totalsize /nBatches )); 
    %actual number of samples, may be less than SamplesPerBatch
        
    if nBatches > 1
        xxTrain = batchdata_reshape( xxTrain2Data{c}(:,1:nBatches*nSamples)', [nSamples d nBatches]);
    else
        xxTrain = xxTrain2Data{c}(:,1:nBatches*nSamples)';
    end
    clear paramsmfa; paramsmfa.PreWts =[];
    paramsmfa.PreWts.Psi = 1*ones(d, C2(c) );    
    paramsmfa.PreWts.M = 0.01*randn(d, C2(c) );
    paramsmfa.PreWts.W = 0.01*randn( d, d2, C2(c) );
    
    paramsmfa.max_iters = 100;
    paramsmfa.num_mbs = min(nBatches, 10);
    
    paramsmfa.GPU = 0;
    paramsmfa.iso = 0;
    paramsmfa.eq = 0;
    paramsmfa.QTHRESH = 1e-5;
    paramsmfa.MIN_PSI = 1e-10;
    paramsmfa.epsilon = 1e-4;
    paramsmfa.ZM = 0;
    
    [LogL, mfalayer2{c}] = mfa( d2, C2(c) , paramsmfa);
    
    if c <= length(xxTest2Data) && ~isempty(xxTest2Data{c})
        test_mean_logprob_init = loggausspdf( double(xxTest2Data{c}), zeros(d,1), eye(d));
        likv = eval_mfa( double(xxTest2Data{c}), mfalayer2{c} );
        test_mean_logprob = mean(likv);
    
        fprintf('======c:%d/%d test:%.3f %.3f ======\n\n\n', c,C, mean(test_mean_logprob_init), mean(test_mean_logprob));
    end
end

clear xxTrain2;
clear xxTrain2Data;

%% convert to a shallow MFA (has non EYE covariance on z)
mfacollapsed =[]; 

C1 = size(mfamodel2.W,3);
C2_max = max(C2);

latdim2 = size(mfalayer2{1}.W,2);
latdim1 = size(mfamodel2.W,2);
pr = zeros(C1, C2_max);

mfacollapsed.M = [];
mfacollapsed.Psi1 = [];
mfacollapsed.Sigm = [];
mfacollapsed.W1 = [];
mfacollapsed.pr = [];

ind = 1;
for i = 1:C1    
    for j = 1:C2(i)
        
       pr(i,j) = mfamodel2.mix(i)*mfalayer2{i}.mix(j);
       mfacollapsed.pr(ind) = pr(i,j);
       %mfa_shallow.WW2(:,:, ind) = mfamodel2.W(:,:,i)*mfalayer2{i}.W(:,:,j);
       mfacollapsed.M(:, ind) = mfamodel2.W(:,:,i)*mfalayer2{i}.M(:,j)+mfamodel2.M(:,i);       
       mfacollapsed.Psi1(:, ind) = mfamodel2.Psi(:,i);
       mfacollapsed.Sigm(:,:,ind) = diag(mfalayer2{i}.Psi(:,j))+mfalayer2{i}.W(:,:,j)*mfalayer2{i}.W(:,:,j)';
       mfacollapsed.W1(:,:,ind) = mfamodel2.W(:,:,i);
       %mfa_shallow.W2(:,:,ind) = mfalayer2{i}.W(:,:,j);
       ind = ind+1;
    end
end


%% convert to a shallow MFA with identity cov matrix for z
mfacollapsed2 = [];

C1 = size(mfamodel2.W,3);
C2_max = max(C2);

latdim2 = size(mfalayer2{1}.W,2);
latdim1 = size(mfamodel2.W,2);
pr = zeros(C1, C2_max);


mfacollapsed2.W = [];
mfacollapsed2.M = [];
mfacollapsed2.Psi = [];
mfacollapsed2.mix = [];

ind = 1;
for i = 1:C1
    for j = 1:C2(i)
        
       pr(i,j) = mfamodel2.mix(i)*mfalayer2{i}.mix(j);
       mfacollapsed2.mix(ind,1) = pr(i,j);
       
       Sigma = diag(mfalayer2{i}.Psi(:,j))+mfalayer2{i}.W(:,:,j)*mfalayer2{i}.W(:,:,j)';
       L = chol(Sigma, 'lower');
       mfacollapsed2.W(:,:,ind) =  mfamodel2.W(:,:,i)*L;           
       mfacollapsed2.M(:, ind) = mfacollapsed2.W(:,:,ind)*(L \ mfalayer2{i}.M(:,j))...
                                +mfamodel2.M(:,i);       
       mfacollapsed2.Psi(:, ind) = mfamodel2.Psi(:,i);      
       ind = ind+1;
    end
end


likv = eval_mfa( double(xxTest'), mfamodel2);
mfamodel2_testlogp = mean(likv);

LogL = ll_deep_mfa(double(xxTest'), mfacollapsed ,1);
likv = logsumexp(bsxfun(@plus, LogL, log(mfacollapsed.pr')), 1);
mfa_shallow_testlogp = mean(likv);

%test log probability
likv = eval_mfa( double(xxTest'), mfacollapsed2);
mfa_shallow2_testlogp = mean(likv);


fprintf('\n shallow mfa evalu logprob: orig:%f collapse:%f collapse2:%f\n' ,...
   mfamodel2_testlogp, mfa_shallow_testlogp, mfa_shallow2_testlogp );




