%{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
CT 2/2012 - this function turns a 2 layer mfa into a 3 layer mfa 
but will collapse it
PURPOSE:
INPUT:  params_dmfa.nSamples - how many posterior samples do we take per FA
        mfamodel - original 1 layer mfa
        mfalayer2 - cell, 1 for every 1st layer FA.
    
        
        
OUTPUT: mfacollapsed has non EYE covariance on z
        mfalayer3{i,j} - has J mfa for I 2nd layer mfa       
NOTES:          
TESTED:
CHANGELOG:
TODO:

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}
function [mfacollapsed mfalayer3] = deepen_mfa3(params_dmfa, mfamodel, mfalayer2)

global xxTrain;
global xxTest;
global xxValid;
xxValid = [];

assert(length(size(xxTrain)) == 2 && length(size(xxTest)) == 2);

Dim = size(xxTrain,2);
assert(Dim == size(xxTest,2));


%% get data for higher layer
clear xxTrain3Data;
clear xxTrain3;

C1 = size(mfamodel.M,2);
d1 = size(mfamodel.W,2);

nSamples = params_dmfa.nSamples;  %enlarge the dataset by nSamples times

if size(xxTrain,1) > 10000
    xxTrain = batchdata_reshape(xxTrain, [size(xxTrain,1)/10 size(xxTrain,2) 10]);
end

for bb = 1:size(xxTrain,3)    
    fprintf(' %.2f', bb/double(size(xxTrain,3)));
    
    [LogPxGc, Lats, lat_covs] = ll_mfa( double(xxTrain(:,:,bb)'), mfamodel, double(xxTrain(:,:,bb)').^2, 1);   
    LogPxc = bsxfun(@plus, LogPxGc, log(mfamodel.mix+realmin));
    Q    = bsxfun(@minus, LogPxc, max(LogPxc,[],1) );
    Q    = exp(Q)+realmin;
    Q    = bsxfun(@rdivide, Q , sum(Q,1) );
    
    [val Q_which_c] = max(Q, [], 1); %make a hard choice!   
   
    for c = 1:C1
        inds = find(Q_which_c == c);
        assert(~isempty(inds));
        LatsSmaller = Lats(:, inds, c);
        
        h1 = LatsSmaller;
        
        %L = chol(lat_covs(:,:,c), 'lower');        
        %xxTrain2{c,bb} =  xxTrain2{c,bb} + L*randn(d1, nSamples*size(LatsSmaller,2) ); %add some posterior noise
        
        %go up one more layer
        C2 = size(mfalayer2{c}.M,2);
        d2 = size(mfalayer2{c}.W,2);
        
        [LogPxGc2, Lats2, lat_covs2] = ll_mfa( double(h1), mfalayer2{c}, double(h1).^2, 1);
       
        LogPxc2 = bsxfun(@plus, LogPxGc2, log(mfalayer2{c}.mix+realmin));
        Q2    = bsxfun(@minus, LogPxc2, max(LogPxc2,[],1) );
        Q2   = exp(Q2)+realmin;
        Q2    = bsxfun(@rdivide, Q2 , sum(Q2,1) );        
        [val Q_which_c2] = max(Q2, [], 1); %make a hard choice!
        
        for c2 = 1:C2
            inds2 = find(Q_which_c2 == c2);
            assert(~isempty(inds2));
            
            Lats2Smaller = Lats2(:, inds2, c2);            
            xxTrain3{c, c2, bb} = repmat(Lats2Smaller, 1, nSamples);
            
            L = chol(lat_covs2(:,:,c2), 'lower');
            xxTrain3{c, c2, bb}  =  xxTrain3{c, c2, bb}  + L*randn(d2, nSamples*size(Lats2Smaller,2) ); 
            %add some posterior noise
        end
    end    
end

xxTrain = batchdata_reshape(xxTrain);

for c = 1:C1
    for c2 = 1:C2
        xxTrain3Data{c, c2} = cat(2,xxTrain3{c,c2,:});
    end
end

clear Lats;
clear Lats2;

% ----- test data -----
xxTest3=[];
[LogPxGc, Lats, lat_covs] = ll_mfa( double(xxTest'), mfamodel, double(xxTest').^2, 1);
LogPxc = bsxfun(@plus, LogPxGc, log(mfamodel.mix+realmin));
Q    = bsxfun(@minus, LogPxc, max(LogPxc,[],1) );
Q    = exp(Q)+realmin;
Q    = bsxfun(@rdivide, Q , sum(Q,1) );

[val Q_which_c] = max(Q, [], 1); %make a hard choice!

for c = 1:C1
    inds = find(Q_which_c == c);
    if isempty(inds)
        continue;
    end
    
    LatsSmaller = Lats(:, inds, c);    
    h1 = LatsSmaller;    
    %L = chol(lat_covs(:,:,c), 'lower');
    %xxTrain2{c,bb} =  xxTrain2{c,bb} + L*randn(d1, nSamples*size(LatsSmaller,2) ); %add some posterior noise
    
    %go up one more layer
    C2 = size(mfalayer2{c}.M,2);
    d2 = size(mfalayer2{c}.W,2);
    
    [LogPxGc2, Lats2, lat_covs2] = ll_mfa( double(h1), mfalayer2{c}, double(h1).^2, 1);
    
    LogPxc2 = bsxfun(@plus, LogPxGc2, log(mfalayer2{c}.mix+realmin));
    Q2    = bsxfun(@minus, LogPxc2, max(LogPxc2,[],1) );
    Q2   = exp(Q2)+realmin;
    Q2    = bsxfun(@rdivide, Q2 , sum(Q2,1) );
    [val Q_which_c2] = max(Q2, [], 1); %make a hard choice!
    
    for c2 = 1:C2
        inds2 = find(Q_which_c2 == c2);
        if isempty(inds2)
            continue;
        end
        
        Lats2Smaller = Lats2(:, inds2, c2);
        xxTest3{c, c2} = repmat(Lats2Smaller, 1, nSamples);
        
        L = chol(lat_covs2(:,:,c2), 'lower');
        xxTest3{c, c2}  =  xxTest3{c, c2}  + L*randn(d2, nSamples*size(Lats2Smaller,2) );
        %add some posterior noise
    end
end

clear Lats;
clear Lats2;

%% learn higher layer

C3 = params_dmfa.C3;
d2 = size(mfalayer2{1}.W,2);
d3 = params_dmfa.d3;

assert( d3 <= d2);
mfalayer3 = cell(C1,C2);  % cell to store a MFA for each 1st layer FA, and 2nd layer FA

for c = 1:C1
    for c2 = 1:C2
        SamplesPerBatch = 100000;
        totalsize = double(size(xxTrain3Data{c, c2},2));
        nBatches = double(max(1, int32(floor( totalsize/SamplesPerBatch)) ));
        nSamples = int32(floor( totalsize /nBatches )); 
        %actual number of samples, may be less than SamplesPerBatch

        if nBatches > 1
            xxTrain = batchdata_reshape( xxTrain3Data{c, c2}(:,1:nBatches*nSamples)', [nSamples d2 nBatches]);
        else
            xxTrain = xxTrain3Data{c, c2}(:,1:nBatches*nSamples)';
        end
        clear paramsmfa; paramsmfa.PreWts =[];
        paramsmfa.PreWts.Psi = 1*ones(d2, C3);    
        paramsmfa.PreWts.M = 0.001*randn(d2, C3);
        paramsmfa.PreWts.W = 0.001*randn(d2, d2, C3);

        paramsmfa.max_iters = 100;
        paramsmfa.num_mbs = min(nBatches, 10);

        paramsmfa.GPU = 0;
        paramsmfa.iso = 0;
        paramsmfa.eq = 0;
        paramsmfa.QTHRESH = 1e-5;
        paramsmfa.MIN_PSI = 1e-10;
        paramsmfa.epsilon = 1e-4;
        paramsmfa.ZM = 0;

        [LogL, mfalayer3{c, c2}] = mfa( d3, C3, paramsmfa);

        if c <= size(xxTest3, 1) && c2<=size(xxTest3,2) && ~isempty(xxTest3{c,c2})
            test_mean_logprob_init = loggausspdf( double(xxTest3{c,c2}), zeros(d2,1), eye(d2));
            likv = eval_mfa( double(xxTest3{c,c2}), mfalayer3{c,c2} );
            test_mean_logprob = mean(likv);

            fprintf('======c1:%d/%d c2:%d/%d test:%.3f %.3f ======\n\n\n', c,C1, c2, C2, mean(test_mean_logprob_init), mean(test_mean_logprob));
        end
    end
end

clear xxTrain3;
clear xxTrain3Data;


%% convert to a shallow MFA with identity cov matrix for z
%collapse onto 1 layer
mfacollapsed = [];

C1 = size(mfamodel.W,3);
C2 = size(mfalayer2{1}.W,3);
C3 = size(mfalayer3{1}.W,3);

latdim1 = size(mfamodel.W,2);
pr = zeros(C1, C2);


mfacollapsed.W = zeros( Dim, latdim1, C1*C2*C3);
mfacollapsed.M = zeros( Dim, C1*C2*C3);
mfacollapsed.Psi = zeros( Dim, C1*C2*C3);
mfacollapsed.mix = zeros(C1*C2*C3, 1);

ind = 1;
for i = 1:C1
    for j = 1:C2
        for k = 1:C3
            pr(i,j,k) = mfamodel.mix(i)*mfalayer2{i}.mix(j)*mfalayer3{i,j}.mix(k);
                       
            mfacollapsed.mix(ind) = pr(i,j,k);
            
            Sigma2 = diag(mfalayer3{i,j}.Psi(:,k))+mfalayer3{i,j}.W(:,:,k)*mfalayer3{i,j}.W(:,:,k)';
            Sigma = diag(mfalayer2{i}.Psi(:,j))+mfalayer2{i}.W(:,:,j)*Sigma2*mfalayer2{i}.W(:,:,j)';
            L = chol(Sigma, 'lower');
            mfacollapsed.W(:,:,ind) =  mfamodel.W(:,:,i)*L;
            mfacollapsed.M(:,ind) = mfamodel.W(:,:,i)*( mfalayer2{i}.W(:,:,j)*mfalayer3{i,j}.M(:,k)+mfalayer2{i}.M(:,j)) ...
                                     + mfamodel.M(:,i);
            mfacollapsed.Psi(:, ind) = mfamodel.Psi(:,i);
            
            ind = ind+1;
        end
    end
end

likv = eval_mfa( double(xxTest'), mfamodel);
mfamodel2_testlogp = mean(likv);

likv = eval_mfa( double(xxTest'), mfacollapsed);
mfa_shallow_testlogp = mean(likv);

fprintf('\n Deepen3 shallow mfa evalu logprob: orig:%f collapse:%f \n' ,...
   mfamodel2_testlogp, mfa_shallow_testlogp );




