%{
===========================================================================
Code provided by Yichuan (Charlie) Tang
http://www.cs.toronto.edu/~tang

Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.


This file is the main script for training a deep MFA on cifar images
===========================================================================
%}

clear;
rand('seed', 1785);
addpath './util';
addpath './mfa';


batchdata1 = load( 'data_batch_1.mat');
test_batch = load( 'test_batch.mat');

Data = single([batchdata1.data;])/255;
TestX = single(test_batch.data)/255;

clear batchdata1;
clear test_batch;


Data = bsxfun(@minus, Data, mean(Data, 2) );
TestX = bsxfun(@minus, TestX,  mean(TestX, 2) );


%% test MoFA

global xxTrain;
global xxTest;
global xxValid;

%only uses 3000 for training here!
xxTrain = Data(1:100*30,:); % NxD
xxValid = Data(100*50+1:100*50+1000,:); % NxD

xxTest = TestX;   % NxD
xxTrain0 = xxTrain;

clear Data;
clear TestX;
 

%%  Train basic MFA with 50 components

params_bench.GPU = zeros(1,100);
params_bench.trials = 1;
params_bench.max_epochs = 1;
params_bench.max_iters = 10; %number of iterations per epoch

params_bench.RandInitFactor = .01;
params_bench.PsiInitFactor = 100;
%maybe need to be 10 or greater for learning stability

params_bench.KMEANS = 2;
params_bench.EQ = 0;
params_bench.save_interm_wts = 'dmfa_cifar.mat';  %filename to save to
params_bench.QTHRESH = 1e-5;
params_bench.MIN_PSI = 1e-8;
params_bench.epsilon = 1e-4;

params_bench.Kvec = 30; % Number of factors per mixture
params_bench.numparams = params_bench.Kvec*11;
params_bench.ZM = 0; %enforce zero mean
params_bench.LOAD_PREV_WTS = 0;

[mfa_bm_res mfamodel] = mfa_benchmark( params_bench);
validLogL = cat(1, mfa_bm_res(:).logl_test);
[v ind] = max(validLogL);
mfa_bm_res(ind);



%% Train a second and third layer layer deep model

for trials = 1:2
    
    % Train a second layer deep model
    xxTrain = xxTrain0;
    params_dmfa =[];
    params_dmfa.nSamples = 5;
    params_dmfa.prune = 0.001;
    params_dmfa.C2 =  5;
    params_dmfa.d2 = 20;
    params_dmfa.smart_C2 =0;

    [mfacollapsed mfacollapsed2 mfamodel2 mfalayer2] = deepen_mfa( mfamodel, params_dmfa);

    likv = eval_mfa( double(xxTrain0'), mfamodel);
    trainll(trials, 1) = mean(likv);
    likv = eval_mfa( double(xxTest'), mfamodel);
    testll(trials, 1) = mean(likv);

    %train log probability
    likv = eval_mfa( double(xxTrain0'), mfacollapsed2);
    mfacollapsed2_trainlogp = mean(likv);
    likv = eval_mfa( double(xxTest'), mfacollapsed2);
    mfacollapsed2_testlogp = mean(likv);

    trainll(trials, 2) = mfacollapsed2_trainlogp;
    testll(trials, 2) = mfacollapsed2_testlogp;
    
    % Train a third layer deep model
    xxTrain = xxTrain0;
    params_dmfa3.C3 = 3;
    params_dmfa3.d3 = 15;
    params_dmfa3.nSamples = 20;
    [mfacollapsed3 mfalayer3] = deepen_mfa3(params_dmfa3, mfamodel2, mfalayer2);

    likv = eval_mfa( double(xxTrain0'), mfacollapsed3);
    mfacollapsed3_trainlogp = mean(likv)
    likv = eval_mfa( double(xxTest'), mfacollapsed3);
    mfacollapsed3_tt = mean(likv)

    trainll(trials, 3) = mfacollapsed3_trainlogp;
    testll(trials, 3) = mfacollapsed3_tt;

end

mfa2_diff_train = trainll(:,2)-trainll(:,1);
mfa2_diff_test = testll(:,2)-testll(:,1);
mfa3_diff_train = trainll(:,3)-trainll(:,2);
mfa3_diff_test = testll(:,3)-testll(:,2);

Trials = size(trainll,1);
fprintf('\nTrain: mfa2 mean:%f stderr:%f mfa3 mean:%f stderr:%f', ...
    mean(mfa2_diff_train), std(mfa2_diff_train)/sqrt(Trials),...
    mean(mfa3_diff_train), std(mfa3_diff_train)/sqrt(Trials) );
fprintf('\nTest: mfa2 mean:%f stderr:%f mfa3 mean:%f stderr:%f', ...
    mean(mfa2_diff_test), std(mfa2_diff_test)/sqrt(Trials),...
    mean(mfa3_diff_test), std(mfa3_diff_test)/sqrt(Trials) );




