%{

% CT adapted ll_mfa for deep mfa (which is converted back to a shallow net)
%
% computes log-likelihoods of X under Gaussians
% also returns the inferred latent coordinates
%
% X (D x N)     - n d-dimensional data
% mfa is a struct with the fields:
% M (D x C)     - k d-dimensional means
% W (D x d x C) - for each Gaussian q d-dim vectors
% Psi (D,C)     - diagonal noise levels

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================
%}
function [ LogL ] = ll_deep_mfa( X, deep_mfa, NEWFORMULATION)


N = size(X,2);
[D, C] = size(deep_mfa.M);

LogL = zeros(C,N);

fprintf('\n');
for c = 1:C
    fprintf('+');
    W1             = deep_mfa.W1(:,:,c);
    Psi           = deep_mfa.Psi1(:,c);
    PsiI          = Psi.^-1;    
    Vc =       inv(deep_mfa.Sigm(:,:,c))+bsxfun(@times, W1', PsiI')*W1;
    
    PsiIw         = bsxfun(@times, W1, PsiI);
    PsiI_w_xc     = PsiIw' * X - repmat(PsiIw'*deep_mfa.M(:,c),1,N);
    
    if NEWFORMULATION ==1
        L=chol(Vc);
        log_det_Vc = 2*sum(log(diag(L)));
    else
        warning('ll_deep_mfa, NEWFORMULATION == 0');
        VcInv = inv(Vc);
        L=chol(VcInv);
        log_det_Vc = -2*sum(log(diag(L)));
    end
    
    det_sigm = det(deep_mfa.Sigm(:,:,c));
    if 1 || det_sigm == 0
        L=chol(deep_mfa.Sigm(:,:,c));
        log_det_sigm = 2*sum(log(diag(L)));
    else
        log_det_sigm = log( det_sigm );
    end
    
    %calculate log determinant of the covariance matrix of p(x)    
    log_det = sum(log(Psi))  + log_det_sigm + log_det_Vc; %equation 12, verbeek ieee paper
        
    if NEWFORMULATION ==1
        energy  =  -.5*( PsiI'*(X.^2) + PsiI'*(deep_mfa.M(:,c).^2)-2*(PsiI.*deep_mfa.M(:,c))'*X ) ...
                + .5 * sum(  (PsiI_w_xc'/Vc) .* PsiI_w_xc', 2)';
    else
        energy  =  -.5*( PsiI'*(X.^2) + PsiI'*(deep_mfa.M(:,c).^2)-2*(PsiI.*deep_mfa.M(:,c))'*X ) ...
                + .5 * sum(  (PsiI_w_xc'*VcInv) .* PsiI_w_xc', 2)';
    end
    
    %assert( log_det_lat_cov + log_det_Vc < 0.01);
    %maxabs( PsiI_w_xc'*VcInv,  PsiI_w_xc'/Vc)
    
    LogL(c,:)     = energy -.5*log_det -(D/2)*log(2*pi);    
end



