%{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
CT 2/12
PURPOSE:   E-step
INPUT:     X - data
           X2 - precomputed X.^2           

       Usage: [Q, LogL, Lats, lat_covs] = mfa_E_step(X,mfa,X2); % E-step

OUTPUT:
NOTES:          
TESTED:
CHANGELOG: modified from original verbeek code
TODO:

% Originally by Jakob Verbeek, Intelligent Systems Laboratory Amsterdam, 
% University of Amsterdam, 2005.
% http://www.science.uva.nl/~jverbeek

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}
function [Q, LogL, Lats, lat_covs] = mfa_E_step( X, mfa, X2 )


N       = size(X,2);
[D,d,C] = size(mfa.W);

[LogPxGc, Lats, lat_covs] = ll_mfa(X, mfa, X2);
assert(all(isfinite(LogPxGc(:))));

LogPxc = bsxfun(@plus, LogPxGc, log(mfa.mix+realmin));

Q    = bsxfun(@minus, LogPxc, max(LogPxc,[],1) );
Q    = exp(Q)+realmin;
Q    = bsxfun(@rdivide, Q , sum(Q,1) );

%assert( abs(sum(Q(:)) - size(Q,2) ) < 1e-5 );

LogL = sum(Q.*(-log(Q+realmin) + LogPxc),1); 
%lower bound (which is tight, since Q is exactly the posterior)
assert(all(isfinite(LogL)));

