%{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
CT 2/12
PURPOSE:   M step update
INPUT:     params.iso - for each mixture component have equal Psi(i)
           params.eq  - all components have same Psi
           params.GPU - 0 use CPU, 1 use GPU for matrix multiplication
           
           params.QTHRESH - if the average (over samples) posterior 
                    probability of a component is less than this thresh
                    we don't update the parameters of that component
           params.MIN_PSI - no Psi value can be less than this             
OUTPUT:
NOTES:          
TESTED:
CHANGELOG: modified from original verbeek code
TODO:

% Originally by Jakob Verbeek, Intelligent Systems Laboratory Amsterdam, 
% University of Amsterdam, 2005.
% http://www.science.uva.nl/~jverbeek

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}

function [mfa] = mfa_M_step( params, X, X2, Q, Lats, lat_covs, oldmfa, ZM)

D       = size(X,1);
[d,N,C] = size(Lats);

if ~exist('X2', 'var') || isempty(X2)
    X2 = X.^2;
end

if ~exist('ZM', 'var') || isempty(ZM)
    ZM = 0;
end

% first copy over the previous parameters
mfa.W = oldmfa.W;
mfa.M = oldmfa.M;
mfa.Psi = oldmfa.Psi;
mfa.mix = oldmfa.mix;

for c=1:C; % M-step
        sumQ         = sum(Q(c,:));
        
        if sumQ/N > params.QTHRESH   %update the parameters
            
            lats         = [Lats(:,:,c); ones(1,N)];
            Qlats        = bsxfun(@times, Q(c,:), lats);

            if params.GPU == 0 %Slow
                XQlats       = X*Qlats'; 
                tmp          = lats*Qlats';                
            else %GPU        
                XQlats       = double( mexcuMatMul( single(X), single(Qlats')) ); 
                tmp          = double( mexcuMatMul( single(lats), single(Qlats')) );
            end
            
            tmp(1:d,1:d) = tmp(1:d,1:d) + sumQ * lat_covs(:,:,c);
            tmp          = mrdivide( XQlats  ,  tmp);

            mfa.W(:,:,c)     = tmp(:,1:d);
            
            mfa.M(:,c)       = tmp(:,d+1);
            
            if params.GPU == 0 %Slow
                mfa.Psi(:,c)     = ( X2*Q(c,:)' - sum(tmp.*XQlats,2))/sumQ;
            else
                mfa.Psi(:,c)     = (double(mexcuMatMul(single(X2), single(Q(c,:)'))) - sum(tmp.*XQlats,2))/sumQ;            
            end
            
            mfa.Psi(:,c)     = max(mfa.Psi(:,c), params.MIN_PSI );
            
            if params.iso==1
                mfa.Psi(:,c) = mean(mfa.Psi(:,c));
            end
        end
        
        mfa.mix(c,1)     = sumQ / N;
end

if params.eq==1
    mfa.Psi = (mfa.Psi*mfa.mix)*ones(1,C); 
end

% prior: give some components a little bit of mixture
smallmix = 0.1/C;
mfa.mix = mfa.mix+smallmix;
mfa.mix = mfa.mix/sum(mfa.mix);

if ZM == 1  %enforce that mixture component means must be zero
    mfa.M(:) = 0;
end

