%{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
This is a wrapper to test the mfa as a generative model
CT 10/2011
PURPOSE:
INPUT:   params.numparams - desired total number of parameters
         params.Kvec - a vector of K (factors) for each trial
         params.EQ - 0 means all component can have different Psi
                     1 means they all have to have equal Psi
         params.KMEANS - 0 means initialize w/ random training points
                       - 1 means run 10 iteration so kmeans
                       - 2 small jitter from mean of complete data

         params.RandInitFactor = .01 (good default);
         params.PsiInitFactor = .05   (good default);
         params.max_iters - maximum number of iterations
         params.trials - how many different trials required
         params.GPU - use GPU or not
         params.save_interm_wts - file name to save intermediate learned mfa models
            
         xxTrain, xxTest should be global, and are nsamples by dim by
         batches

OUTPUT:
NOTES:
TESTED:
CHANGELOG:
TODO:

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}
function [results mfamodel] = mfa_benchmark( params )

ind = 1;
results = [];

global xxTrain;
global xxTest;

[nsamples dim batches] = size(xxTrain);
    
for K = double(params.Kvec)
    
    if batches > 1
        xxTrain = batchdata_reshape(xxTrain);
    end
    
    M = round(params.numparams/(K+2)); %approx satisfies total number of parameters
    if params.KMEANS == 1        
        [IDX Cntrs] = kmeans(xxTrain(1:100000,:), M, 'MaxIter', 20, 'Display', 'iter');        
    elseif params.KMEANS == 0
        inds = randperm(size(xxTrain,1));
        Cntrs = single(xxTrain(inds(1:M),:));
    elseif params.KMEANS == 2
        Cntrs = repmat( mean(xxTrain), [M, 1]);
        Cntrs = Cntrs + 0.01*randn(size(Cntrs));
    end
    
    paramsmfa = [];
    paramsmfa.PreWts.Psi = repmat( params.PsiInitFactor*(var(xxTrain, 0, 1)'), 1, M);
    paramsmfa.PreWts.M = Cntrs';
    paramsmfa.PreWts.W = params.RandInitFactor*randn( size(xxTrain,2) , K, M);
        
    if batches > 1
        xxTrain = batchdata_reshape(xxTrain, [nsamples dim batches]);        
    end
    
    paramsmfa.max_iters = params.max_iters;   
    paramsmfa.ZM = params.ZM;
    MoFA = [];
    
    if params.LOAD_PREV_WTS ==1
       dd = load(params.prev_wts_fn);
       paramsmfa.PreWts.Psi = dd.MoFA{end}.Psi;
       paramsmfa.PreWts.M = dd.MoFA{end}.M;
       paramsmfa.PreWts.W =dd.MoFA{end}.W;
       paramsmfa.PreWts.mix =dd.MoFA{end}.mix;
    end       
    
    % starting training epochs
    for ee = 1:params.max_epochs
                
        paramsmfa.num_mbs = 1;
        
        paramsmfa.GPU = 0;
        paramsmfa.iso = 0;
        paramsmfa.eq = params.EQ;
        paramsmfa.QTHRESH = params.QTHRESH;
        paramsmfa.MIN_PSI = params.MIN_PSI;
        paramsmfa.epsilon = params.epsilon;
        [LL, mfamodel, Q] = mfa( K, M, paramsmfa);
        
        paramsmfa.PreWts.Psi = mfamodel.Psi;
        paramsmfa.PreWts.M = mfamodel.M;
        paramsmfa.PreWts.W = mfamodel.W;
        paramsmfa.PreWts.mix = mfamodel.mix;
                
        likv = eval_mfa( double(xxTest'), mfamodel);
        test_mean_logprob = mean(likv);
            
        fprintf('\n====== ee:%d test:%f ======', ee, test_mean_logprob);
        
        MoFA{ee} = mfamodel;
        save( params.save_interm_wts, 'MoFA', 'ee');
    
    end % max_epochs
    
    
    results(ind).K = K;
    results(ind).M = M;
    %results(ind).logl_train =  mean(train_mean_logprob);
    results(ind).logl_test =  mean(test_mean_logprob);
    results(ind).totalparams = (K+2)*M+M;
    
    fprintf('\n =================== %d =====================\n', ind);
    ind = ind + 1;

end % Kvec
